import shlex
import sys
from typing import Optional

import click
from rich.console import Console

from hapless import config
from hapless.main import Hapless

console = Console(highlight=False)
hapless = Hapless()


def get_or_exit(hap_alias: str):
    hap = hapless.get_hap(hap_alias)
    if hap is None:
        console.print(
            f"{config.ICON_INFO} No such hap: {hap_alias}",
            style=f"{config.COLOR_ERROR} bold",
        )
        sys.exit(1)
    return hap


@click.group(invoke_without_command=True)
@click.version_option(message="hapless, version %(version)s")
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        _status()


@cli.command()
@click.argument("hap_alias", metavar="hap", required=False)
def status(hap_alias):
    _status(hap_alias)


@cli.command()
@click.argument("hap_alias", metavar="hap", required=False)
def show(hap_alias):
    _status(hap_alias)


def _status(hap_alias: Optional[str] = None):
    if hap_alias is not None:
        hap = get_or_exit(hap_alias)
        hapless.show(hap)
    else:
        haps = hapless.get_haps()
        hapless.stats(haps)


@cli.command()
@click.argument("hap_alias", metavar="hap")
@click.option("-f", "--follow", is_flag=True, default=False)
def logs(hap_alias, follow):
    hap = get_or_exit(hap_alias)
    hapless.logs(hap, follow=follow)


@cli.command()
@click.option("--skip-failed", is_flag=True, default=False)
def clean(skip_failed):
    hapless.clean(skip_failed)


@cli.command()
@click.argument("cmd", nargs=-1)
@click.option("--check", is_flag=True, default=False)
def run(cmd, check):
    hapless.run(shlex.join(cmd), check=check)


if __name__ == "__main__":
    cli()



import asyncio
import os
import shutil
import subprocess
import sys
import tempfile

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    # Fallback for Python 3.7
    from importlib_metadata import version
from pathlib import Path
from typing import Dict, List, Optional

from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from hapless import config
from hapless.hap import Hap
from hapless.utils import wait_created

"""
paused
running
finished
* finished(failed) non-zero rc
* finished(success) zero rc
💀
"""
console = Console(highlight=False)


class Hapless(object):
    def __init__(self):
        tmp_dir = Path(tempfile.gettempdir())
        self._hapless_dir = tmp_dir / "hapless"
        if not self._hapless_dir.exists():
            self._hapless_dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def stats(haps: List[Hap]):
        if not haps:
            console.print(
                f"{config.ICON_INFO} No haps are currently running",
                style=f"{config.COLOR_MAIN} bold",
            )
            return

        package_version = version(__package__)
        table = Table(
            title=f"{config.ICON_HAP} {__package__}, {package_version}",
            show_header=True,
            header_style=f"{config.COLOR_MAIN} bold",
            box=box.HEAVY_EDGE,
        )
        table.add_column("#", style="dim", width=2)
        table.add_column("Name")
        table.add_column("PID")
        table.add_column("Status")
        table.add_column("RC")
        table.add_column("Runtime", justify="right")

        for hap in haps:
            table.add_row(
                f"{hap.hid}",
                hap.name,
                f"{hap.pid}",
                hap.status,
                f"{hap.rc}" if hap.rc is not None else "",
                hap.runtime,
            )

        console.print(table)

    @staticmethod
    def show(hap: Hap):
        table = Table(show_header=False, show_footer=False, box=box.SIMPLE)

        table.add_row("Status:", hap.status)

        table.add_row("PID:", f"{hap.pid}")

        if hap.rc is not None:
            table.add_row("Return code:", f"{hap.rc}")

        cmd_text = Text(f"{hap.cmd}", style=f"{config.COLOR_ACCENT} bold")
        table.add_row("Command:", cmd_text)

        table.add_row("Runtime:", f"{hap.runtime}")

        panel = Panel(
            table,
            expand=False,
            title=f"Hap {config.ICON_HAP}{hap.hid}",
            subtitle=hap.name,
        )
        console.print(panel)

    @property
    def dir(self) -> Path:
        return self._hapless_dir

    def _get_hap_dirs(self) -> List[str]:
        hap_dirs = filter(str.isdigit, os.listdir(self._hapless_dir))
        return sorted(hap_dirs, key=int)

    def _get_hap_names_map(self) -> Dict[str, str]:
        names = {}
        for dir in self._get_hap_dirs():
            filename = self._hapless_dir / dir / "name"
            if filename.exists():
                with open(filename) as f:
                    name = f.read().strip()
                    names[name] = dir
        return names

    def get_next_hap_id(self):
        dirs = self._get_hap_dirs()
        return 1 if not dirs else int(dirs[-1]) + 1

    def get_hap(self, hap_alias: str) -> Optional[Hap]:
        dirs = self._get_hap_dirs()
        # Check by hap id
        if hap_alias in dirs:
            return Hap(self._hapless_dir / hap_alias)

        # Check by hap name
        names_map = self._get_hap_names_map()
        if hap_alias in names_map:
            return Hap(self._hapless_dir / names_map[hap_alias])

    def get_haps(self) -> List[Hap]:
        haps = []
        if not self._hapless_dir.exists():
            return haps

        for dir in self._get_hap_dirs():
            hap_path = self._hapless_dir / dir
            haps.append(Hap(hap_path))
        return haps

    def create_hap(self, cmd: str, name: Optional[str] = None) -> Hap:
        hid = self.get_next_hap_id()
        hap_dir = self._hapless_dir / f"{hid}"
        hap_dir.mkdir()
        return Hap(hap_dir, cmd=cmd, name=name)

    async def run_hap(self, hap: Hap):
        with open(hap.stdout_path, "w") as stdout_pipe, open(
            hap.stderr_path, "w"
        ) as stderr_pipe:
            # todo: run with exec
            proc = await asyncio.create_subprocess_shell(
                hap.cmd,
                stdout=stdout_pipe,
                stderr=stderr_pipe,
            )
            with open(hap._pid_file, "w") as pid_file:
                pid_file.write(f"{proc.pid}")

            pid_text = Text(f"{proc.pid}", style=f"{config.COLOR_MAIN} bold")
            name_text = Text(f"{hap.name}", style=f"{config.COLOR_MAIN} bold")
            console.print(
                f"{config.ICON_INFO} Running hap {config.ICON_HAP}{hap.hid} ",
                "(",
                name_text,
                ") with PID [",
                pid_text,
                "]",
                sep="",
            )
            _ = await proc.communicate()

            with open(hap._rc_file, "w") as rc_file:
                rc_file.write(f"{proc.returncode}")

    def _check_fast_failure(self, hap: Hap):
        if wait_created(hap._rc_file) and hap.rc != 0:
            console.print(
                f"{config.ICON_INFO} Hap exited too quickly. stderr message:",
                style=f"{config.COLOR_ERROR} bold",
            )
            with open(hap.stderr_path) as f:
                console.print(f.read())
            sys.exit(1)

    def run(self, cmd: str, check: bool = False):
        hap = self.create_hap(cmd=cmd)
        pid = os.fork()
        if pid == 0:
            coro = self.run_hap(hap)
            asyncio.run(coro)
        else:
            if check:
                self._check_fast_failure(hap)
            sys.exit(0)

    def logs(self, hap: Hap, follow: bool = False):
        if follow:
            return subprocess.run(["tail", "-f", hap.stdout_path])
        else:
            return subprocess.run(["cat", hap.stdout_path])

    def clean(self, skip_failed: bool = False):
        def to_clean(hap):
            if hap.rc is not None:
                return hap.rc == 0 or not skip_failed
            return False

        haps = list(filter(to_clean, self.get_haps()))
        for hap in haps:
            shutil.rmtree(hap.path)

        if haps:
            console.print(
                f"{config.ICON_INFO} Deleted {len(haps)} finished haps",
                style=f"{config.COLOR_MAIN} bold",
            )
        else:
            console.print(
                f"{config.ICON_INFO} Nothing to clean",
                style=f"{config.COLOR_ERROR} bold",
            )



import os
import random
import string
import time
from pathlib import Path
from typing import Optional

import humanize
import psutil

from hapless import config
from hapless.utils import allow_missing


class Hap(object):
    def __init__(
        self,
        hap_path: Path,
        *,
        name: Optional[str] = None,
        cmd: Optional[str] = None,
    ):
        self._hap_path = hap_path
        self._hid = os.path.basename(hap_path)

        self._pid_file = hap_path / "pid"
        self._rc_file = hap_path / "rc"
        self._name_file = hap_path / "name"
        self._cmd_file = hap_path / "cmd"
        self._env_file = hap_path / "env"

        self._set_name(name)
        self._set_cmd(cmd)
        self._set_env()

    def _set_name(self, name: Optional[str]):
        """
        Sets name for the first time on hap creation.
        """
        if name is None:
            suffix = self.get_random_name()
            name = f"hap-{suffix}"

        if self.name is None:
            with open(self._name_file, "w") as f:
                f.write(name)

    def _set_cmd(self, cmd: Optional[str]):
        """
        Sets cmd for the first tiem on hap creation.
        """
        if self.cmd is None:
            if cmd is None:
                raise ValueError("Command to run is not provided")
            with open(self._cmd_file, "w") as f:
                f.write(cmd)

    def _set_env(self):
        pass

    @staticmethod
    def get_random_name(length: int = 6):
        return "".join(
            random.sample(
                string.ascii_lowercase + string.digits,
                length,
            )
        )

    # todo: add extended status to show panel proc.status()
    @property
    def status(self) -> str:
        proc = self.proc
        # todo: running or paused
        if proc is not None:
            return f"{config.ICON_RUNNING} running"

        if self.rc != 0:
            return f"{config.ICON_FAILED} failed"

        return f"{config.ICON_SUCCESS} success"

    @property
    def proc(self):
        try:
            return psutil.Process(self.pid)
        except psutil.NoSuchProcess:
            pass

    @property
    @allow_missing
    def cmd(self) -> str:
        # todo: might be better for the shell expansion
        # shlex.join(proc.cmdline())
        with open(self._cmd_file) as f:
            return f.read()

    @property
    @allow_missing
    def rc(self) -> Optional[int]:
        with open(self._rc_file) as f:
            return int(f.read())

    @property
    def runtime(self) -> str:
        proc = self.proc
        if proc is not None:
            runtime = time.time() - proc.create_time()
        else:
            start_time = os.path.getmtime(self._pid_file)
            finish_time = os.path.getmtime(self._rc_file)
            runtime = finish_time - start_time

        return humanize.naturaldelta(runtime)

    @property
    def active(self) -> bool:
        return self.proc is not None

    @property
    def hid(self) -> int:
        return self._hid

    @property
    @allow_missing
    def pid(self) -> Optional[int]:
        with open(self._pid_file) as f:
            return int(f.read())

    @property
    @allow_missing
    def env(self):
        pass

    @property
    @allow_missing
    def name(self) -> Optional[str]:
        with open(self._name_file) as f:
            return f.read().strip()

    @property
    def path(self):
        return self._hap_path

    @property
    def stdout_path(self):
        return self._hap_path / "stdout.log"

    @property
    def stderr_path(self):
        return self._hap_path / "stderr.log"

    def __str__(self):
        return f"#{self.hid} ({self.name})"



import time
from functools import wraps


def allow_missing(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except FileNotFoundError:
            pass

    return wrapper


class timed(object):
    def __init__(self):
        self._start = time.time()
        self._end = None

    @property
    def elapsed(self):
        if self._end is not None:
            return self._end - self._start

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._end = time.time()
        return self


def wait_created(path, interval=0.1, timeout=2):
    start = time.time()
    while not path.exists() and time.time() - start < timeout:
        time.sleep(interval)
    return path.exists()



import os

DEBUG = bool(os.getenv("HAPLESS_DEBUG", default=""))
CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

COLOR_MAIN = "#fdca40"
COLOR_ACCENT = "#3aaed8"
COLOR_ERROR = "#f64740"

ICON_HAP = "⚡️"
ICON_INFO = "🧲"
ICON_SUCCESS = "🟢"
ICON_RUNNING = "🟠"
ICON_PAUSED = "⚪️"
ICON_FAILED = "🔴"

FAILFAST_DELAY = 2



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hapless']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.2,<9.0.0',
 'humanize>=4.0.0,<5.0.0',
 'psutil>=5.9.0,<6.0.0',
 'rich>=12.2.0,<13.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.11.3,<5.0.0']}

entry_points = \
{'console_scripts': ['hap = hapless.cli:cli']}

setup_kwargs = {
    'name': 'hapless',
    'version': '0.1.0',
    'description': 'Run and track processes in background',
    'long_description': None,
    'author': 'Misha Behersky',
    'author_email': 'bmwant@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
